// Copyright (c) 2024 The YAC Authors
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <string.h>

#include "basic_grid.h"
#include "grid_cell.h"
#include "field_data_set.h"
#include "io_utils.h"
#include "yac_mpi_internal.h"
#include "time.h"
#include "yac_config.h"
#include "geometry.h"

struct yac_basic_grid {
  char * name;
  int is_empty;
  struct yac_field_data_set * field_data_set;
  struct yac_basic_grid_data data;
};

static struct yac_basic_grid_data yac_basic_grid_data_empty = {
  .vertex_coordinates = NULL,
  .cell_ids = NULL,
  .vertex_ids = NULL,
  .edge_ids = NULL,
  .num_cells = 0,
  .num_vertices = 0,
  .num_edges = 0,
  .core_cell_mask = NULL,
  .core_vertex_mask = NULL,
  .core_edge_mask = NULL,
  .num_vertices_per_cell = NULL,
  .num_cells_per_vertex = NULL,
  .cell_to_vertex = NULL,
  .cell_to_vertex_offsets = NULL,
  .cell_to_edge = NULL,
  .cell_to_edge_offsets = NULL,
  .vertex_to_cell = NULL,
  .vertex_to_cell_offsets = NULL,
  .edge_to_vertex = NULL,
  .edge_type = NULL,
  .num_total_cells = 0,
  .num_total_vertices = 0,
  .num_total_edges = 0
};

struct yac_basic_grid * yac_basic_grid_new(
  char const * name, struct yac_basic_grid_data grid_data) {

  struct yac_basic_grid * grid = xmalloc(1 * sizeof(*grid));

  grid->name = strdup(name);
  grid->is_empty = 0;
  grid->field_data_set = yac_field_data_set_empty_new();
  grid->data = grid_data;

  return grid;
}

struct yac_basic_grid * yac_basic_grid_empty_new(char const * name) {
  struct yac_basic_grid * grid =
    yac_basic_grid_new(name, yac_basic_grid_data_empty);
  grid->is_empty = 1;
  return grid;
}

void yac_basic_grid_delete(struct yac_basic_grid * grid) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_delete): "
    "NULL is not a valid value for argument grid")
  free(grid->name);
  yac_field_data_set_delete(grid->field_data_set);
  yac_basic_grid_data_free(grid->data);
  free(grid);
}

yac_const_coordinate_pointer yac_basic_grid_get_field_coordinates(
  struct yac_basic_grid * grid, struct yac_interp_field field) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_get_field_coordinates): "
    "NULL is not a valid value for argument grid")

  yac_const_coordinate_pointer coords =
    (field.coordinates_idx != SIZE_MAX)?
      yac_field_data_get_coordinates_data(
        yac_basic_grid_get_field_data(grid, field.location),
        field.coordinates_idx):NULL;

  // if no field coordinates are defined, but the location is at the corners of
  // of the grid cells, return coordinates of them
  return
    ((coords != NULL) || (field.location != YAC_LOC_CORNER))?
       coords:((yac_const_coordinate_pointer)(grid->data.vertex_coordinates));
}

int const * yac_basic_grid_get_core_mask(
  struct yac_basic_grid * grid, enum yac_location location) {

  YAC_ASSERT(
    (location == YAC_LOC_CELL) ||
    (location == YAC_LOC_CORNER) ||
    (location == YAC_LOC_EDGE),
    "ERROR(yac_basic_grid_get_core_mask): invalid location")

  switch (location) {
    default:
    case(YAC_LOC_CELL): return grid->data.core_cell_mask;
    case(YAC_LOC_CORNER): return grid->data.core_vertex_mask;
    case(YAC_LOC_EDGE): return grid->data.core_edge_mask;
  };
}

int const * yac_basic_grid_get_field_mask(
  struct yac_basic_grid * grid, struct yac_interp_field field) {

  if (field.masks_idx == SIZE_MAX) return NULL;

  return
    yac_field_data_get_mask_data(
      yac_basic_grid_get_field_data(grid, field.location), field.masks_idx);
}

char const * yac_basic_grid_get_name(struct yac_basic_grid * grid) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_get_name): "
    "NULL is not a valid value for argument grid")

  return grid->name;
}

struct yac_basic_grid_data * yac_basic_grid_get_data(
  struct yac_basic_grid * grid) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_get_data): "
    "NULL is not a valid value for argument grid")

  return &(grid->data);
}

size_t yac_basic_grid_get_data_size(
  struct yac_basic_grid * grid, enum yac_location location) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_get_data_size): "
    "NULL is not a valid value for argument grid")
  YAC_ASSERT(
    (location == YAC_LOC_CELL) ||
    (location == YAC_LOC_CORNER) ||
    (location == YAC_LOC_EDGE),
    "ERROR(yac_basic_grid_get_data_size): invalid location")

  switch (location) {
    default:
    case (YAC_LOC_CELL):
      return grid->data.num_cells;
    case (YAC_LOC_CORNER):
      return grid->data.num_vertices;
    case (YAC_LOC_EDGE):
      return grid->data.num_edges;
  };
}

size_t yac_basic_grid_get_data_size_f2c(
  struct yac_basic_grid * grid, int location) {

  return
    yac_basic_grid_get_data_size(grid, yac_get_location(location));
}

size_t yac_basic_grid_get_named_mask_idx(
  struct yac_basic_grid * grid, enum yac_location location,
  char const * mask_name) {

  if (mask_name == NULL) return SIZE_MAX;

  struct yac_field_data * data =
    yac_basic_grid_get_field_data(grid, location);

  if (data == NULL) return SIZE_MAX;

  size_t mask_idx = SIZE_MAX;
  size_t masks_count = yac_field_data_get_masks_count(data);

  for (size_t i = 0; (i < masks_count) && (mask_idx == SIZE_MAX); ++i) {
    char const * curr_mask_name =
      yac_field_data_get_mask_name(data, i);
    if ((curr_mask_name != NULL) &&
        (!strcmp(mask_name, curr_mask_name)))
      mask_idx = i;
  }

  YAC_ASSERT_F(
    mask_idx != SIZE_MAX,
    "ERROR(yac_basic_grid_get_named_mask_idx): grid \"%s\" does not contain "
    "%s-mask with the name \"%s\"",
    grid->name, yac_loc2str(location), mask_name)

  return mask_idx;
}

size_t yac_basic_grid_add_coordinates_nocpy(
  struct yac_basic_grid * grid,
  enum yac_location location, yac_coordinate_pointer coordinates) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_add_coordinates_nocpy): "
    "NULL is not a valid value for argument grid")
  YAC_ASSERT_F(
    !grid->is_empty, "ERROR(yac_basic_grid_add_coordinates_nocpy): "
    "grid \"%s\" is an empty grid", grid->name)

  return
    yac_field_data_set_add_coordinates_nocpy(
      grid->field_data_set, location, coordinates);
}

size_t yac_basic_grid_add_coordinates_nocpy_f2c(
  struct yac_basic_grid * grid, int location, double * coordinates) {

  return
    yac_basic_grid_add_coordinates_nocpy(
      grid, yac_get_location(location), (yac_coordinate_pointer)coordinates);
}

size_t yac_basic_grid_add_coordinates(
  struct yac_basic_grid * grid, enum yac_location location,
  yac_coordinate_pointer coordinates, size_t count) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_add_coordinates): "
    "NULL is not a valid value for argument grid")
  YAC_ASSERT_F(
    !grid->is_empty, "ERROR(yac_basic_grid_add_coordinates): "
    "grid \"%s\" is an empty grid", grid->name)

  return
    yac_field_data_set_add_coordinates(
      grid->field_data_set, location, coordinates, count);
}

size_t yac_basic_grid_add_coordinates_f2c(
  struct yac_basic_grid * grid, int location,
  double * coordinates, size_t count) {

  return
    yac_basic_grid_add_coordinates(
      grid, yac_get_location(location),
      (yac_coordinate_pointer)coordinates, count);
}

size_t yac_basic_grid_add_mask_nocpy(
  struct yac_basic_grid * grid,
  enum yac_location location, int const * mask,
  char const * mask_name) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_add_mask_nocpy): "
    "NULL is not a valid value for argument grid")
  YAC_ASSERT_F(
    !grid->is_empty, "ERROR(yac_basic_grid_add_mask_nocpy): "
    "grid \"%s\" is an empty grid", grid->name)

  return
    yac_field_data_set_add_mask_nocpy(
      grid->field_data_set, location, mask, mask_name);
}

size_t yac_basic_grid_add_mask_nocpy_f2c(
  struct yac_basic_grid * grid, int location,
  int const * mask, char const * mask_name) {

  return
    yac_basic_grid_add_mask_nocpy(
      grid, yac_get_location(location), mask, mask_name);
}

size_t yac_basic_grid_add_mask(
  struct yac_basic_grid * grid, enum yac_location location,
  int const * mask, size_t count, char const * mask_name) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_add_mask): "
    "NULL is not a valid value for argument grid")
  YAC_ASSERT_F(
    !grid->is_empty, "ERROR(yac_basic_grid_add_mask): "
    "grid \"%s\" is an empty grid", grid->name)

  return
    yac_field_data_set_add_mask(
      grid->field_data_set, location, mask, count, mask_name);
}

size_t yac_basic_grid_add_mask_f2c(
  struct yac_basic_grid * grid, int location,
  int const * mask, size_t count, char const * mask_name) {

  return
    yac_basic_grid_add_mask(
      grid, yac_get_location(location), mask, count, mask_name);
}

struct yac_field_data * yac_basic_grid_get_field_data(
  struct yac_basic_grid * grid, enum yac_location location) {

  YAC_ASSERT(
    grid, "ERROR(yac_basic_grid_get_field_data): "
    "NULL is not a valid value for argument grid")

  if (grid->is_empty)
    return NULL;
  else
    return
      yac_field_data_set_get_field_data(
        grid->field_data_set, location);
}

struct yac_basic_grid * yac_basic_grid_reg_2d_new(
  char const * name, size_t nbr_vertices[2], int cyclic[2],
  double *lon_vertices, double *lat_vertices) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_reg_2d(
        nbr_vertices, cyclic, lon_vertices, lat_vertices));
}

struct yac_basic_grid * yac_basic_grid_reg_2d_deg_new(
  char const * name, size_t nbr_vertices[2], int cyclic[2],
  double *lon_vertices, double *lat_vertices) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_reg_2d_deg(
        nbr_vertices, cyclic, lon_vertices, lat_vertices));
}

struct yac_basic_grid * yac_basic_grid_curve_2d_new(
  char const * name, size_t nbr_vertices[2], int cyclic[2],
  double *lon_vertices, double *lat_vertices) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_curve_2d(
        nbr_vertices, cyclic, lon_vertices, lat_vertices));
}

struct yac_basic_grid * yac_basic_grid_curve_2d_deg_new(
  char const * name, size_t nbr_vertices[2], int cyclic[2],
  double *lon_vertices, double *lat_vertices) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_curve_2d_deg(
        nbr_vertices, cyclic, lon_vertices, lat_vertices));
}

struct yac_basic_grid * yac_basic_grid_unstruct_new(
  char const * name, size_t nbr_vertices, size_t nbr_cells,
  int *num_vertices_per_cell, double *x_vertices, double *y_vertices,
  int *cell_to_vertex) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_unstruct(
        nbr_vertices, nbr_cells, num_vertices_per_cell,
        x_vertices, y_vertices, cell_to_vertex));
}

struct yac_basic_grid * yac_basic_grid_unstruct_deg_new(
  char const * name, size_t nbr_vertices, size_t nbr_cells,
  int *num_vertices_per_cell, double *x_vertices, double *y_vertices,
  int *cell_to_vertex) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_unstruct_deg(
        nbr_vertices, nbr_cells, num_vertices_per_cell,
        x_vertices, y_vertices, cell_to_vertex));
}

struct yac_basic_grid * yac_basic_grid_unstruct_ll_new(
  char const * name, size_t nbr_vertices, size_t nbr_cells,
  int *num_vertices_per_cell, double *x_vertices, double *y_vertices,
  int *cell_to_vertex) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_unstruct_ll(
        nbr_vertices, nbr_cells, num_vertices_per_cell,
        x_vertices, y_vertices, cell_to_vertex));
}

struct yac_basic_grid * yac_basic_grid_unstruct_ll_deg_new(
  char const * name, size_t nbr_vertices, size_t nbr_cells,
  int *num_vertices_per_cell, double *x_vertices, double *y_vertices,
  int *cell_to_vertex) {

  return
    yac_basic_grid_new(
      name,
      yac_generate_basic_grid_data_unstruct_ll_deg(
        nbr_vertices, nbr_cells, num_vertices_per_cell,
        x_vertices, y_vertices, cell_to_vertex));
}

#ifdef YAC_NETCDF_ENABLED

static int def_dim(
  char const * filename, int ncid, char const * name, size_t dim_len,
  int file_is_new) {

  int create_dim = file_is_new;
  int dim_id;

  // if the netcdf file already exists
  if (!file_is_new) {

    // check whether the dimension already exists in the file
    int status = nc_inq_dimid(ncid, name, &dim_id);

    // if the dimension already exists, check for a consistent dimension size
    if (!((create_dim = (status == NC_EBADDIM)))) {

      YAC_HANDLE_ERROR(status);

      size_t temp_dim_len;
      YAC_HANDLE_ERROR(nc_inq_dimlen(ncid, dim_id, &temp_dim_len));

      YAC_ASSERT_F(
        dim_len == temp_dim_len,
        "ERROR(def_dim): file \"%s\" already contains dimension \"%s\", "
        "but it has a different size %zu != %zu",
        filename, name, dim_len, temp_dim_len);
    }
  }

  if (create_dim) YAC_HANDLE_ERROR(nc_def_dim(ncid, name, dim_len, &dim_id));

  return dim_id;
}

static void def_var(
  char const * filename, int ncid, char const * name, nc_type type,
  int ndims, const int * dimids, char const * att_name, char const * att_value,
  int file_is_new) {

  int create_var = file_is_new;
  int var_id;

  // if the netcdf file already exists
  if (!file_is_new) {

    // check whether the variable already exists in the file
    int status = nc_inq_varid(ncid, name, &var_id);

    // if the variable already exists
    if (!((create_var = (status == NC_ENOTVAR)))) {

      YAC_HANDLE_ERROR(status);

      // check the consistency of the variable definition
      nc_type temp_type;
      int temp_ndims;
      int temp_dimids[NC_MAX_VAR_DIMS];
      YAC_HANDLE_ERROR(
        nc_inq_var(
          ncid, var_id, NULL, &temp_type, &temp_ndims, temp_dimids, NULL));
      YAC_ASSERT_F(
        type == temp_type,
        "ERROR(def_var): file \"%s\" already contains variable \"%s\", "
        "but it has a different data type %d != %d",
        filename, name, type, temp_type);
      YAC_ASSERT_F(
        ndims == temp_ndims,
        "ERROR(def_var): file \"%s\" already contains variable \"%s\", "
        "but it has a different number of dimensions %d != %d",
        filename, name, ndims, temp_ndims);
      for (int i = 0; i < ndims; ++i) {
        YAC_ASSERT_F(
          dimids[i] == temp_dimids[i],
          "ERROR(def_var): file \"%s\" already contains variable \"%s\", "
          "but it has a different dimensions dimid[%d] %d != %d",
          filename, name, i, dimids[i], temp_dimids[i])
      }

      if (att_name != NULL) {

        size_t att_len;
        YAC_HANDLE_ERROR(nc_inq_attlen(ncid, var_id, att_name, &att_len));

        char * temp_att_value = xcalloc((att_len + 1), sizeof(temp_att_value));
        YAC_HANDLE_ERROR(nc_get_att_text(ncid, var_id, att_name, temp_att_value));

        YAC_ASSERT_F(
          !strcmp(att_value, temp_att_value),
          "ERROR(def_var): file \"%s\" already contains variable \"%s\", "
          "but it has a different attribute value (name: \"%s\") "
          "\"%s\" != \"%s\"",
          filename, name, att_name, att_value, temp_att_value);
        free(temp_att_value);
      }
    }
  }

  if (create_var) {
    YAC_HANDLE_ERROR(nc_def_var(ncid, name, type, ndims, dimids, &var_id));
    if (att_name != NULL)
      YAC_HANDLE_ERROR(
        nc_put_att_text(ncid, var_id, att_name, strlen(att_value), att_value));
  }
}

static void create_grid_file(
  char const * filename, char const * grid_name, size_t num_cells,
  int num_vertices_per_cell, int cell_center_coords_avaiable,
  int cell_global_ids_available, int core_cell_mask_available) {

  size_t grid_name_len = strlen(grid_name) + 1;

  char nv_dim_name[3 + grid_name_len];
  char nc_dim_name[3 + grid_name_len];

  char cla_var_name[4 + grid_name_len];
  char clo_var_name[4 + grid_name_len];
  char lat_var_name[4 + grid_name_len];
  char lon_var_name[4 + grid_name_len];
  char gid_var_name[4 + grid_name_len];
  char cmk_var_name[4 + grid_name_len];
  char rnk_var_name[4 + grid_name_len];
  char const * unit_att = "units";
  char const * coord_unit = "degree";

  snprintf(nv_dim_name, 3 + grid_name_len, "nv_%s", grid_name);
  snprintf(nc_dim_name, 3 + grid_name_len, "nc_%s", grid_name);

  snprintf(cla_var_name, 4 + grid_name_len, "%s.cla", grid_name);
  snprintf(clo_var_name, 4 + grid_name_len, "%s.clo", grid_name);
  snprintf(lat_var_name, 4 + grid_name_len, "%s.lat", grid_name);
  snprintf(lon_var_name, 4 + grid_name_len, "%s.lon", grid_name);
  snprintf(gid_var_name, 4 + grid_name_len, "%s.gid", grid_name);
  snprintf(cmk_var_name, 4 + grid_name_len, "%s.cmk", grid_name);
  snprintf(rnk_var_name, 4 + grid_name_len, "%s.rnk", grid_name);

  int ncid;

  // create/open file
  int file_is_new = !yac_file_exists(filename);
  if (file_is_new) {
    yac_nc_create(filename, NC_CLOBBER | NC_64BIT_OFFSET, &ncid);
  } else {
    yac_nc_open(filename, NC_WRITE | NC_SHARE, &ncid);
    nc_redef(ncid);
  }

  // define dimensions
  int nv_dim_id =
    def_dim(filename, ncid, nv_dim_name, (size_t)num_vertices_per_cell, file_is_new);
  int nc_dim_id =
    def_dim(filename, ncid, nc_dim_name, (size_t)num_cells, file_is_new);

  // define variables
  int corner_dims[2] = {nc_dim_id, nv_dim_id};
  int cell_dims[1] = {nc_dim_id};
  def_var(
    filename, ncid, cla_var_name, NC_DOUBLE, 2, corner_dims,
    unit_att, coord_unit, file_is_new);
  def_var(
    filename, ncid, clo_var_name, NC_DOUBLE, 2, corner_dims,
    unit_att, coord_unit, file_is_new);
  if (cell_center_coords_avaiable) {
    def_var(
      filename, ncid, lat_var_name, NC_DOUBLE, 1, cell_dims,
      unit_att, coord_unit, file_is_new);
    def_var(
      filename, ncid, lon_var_name, NC_DOUBLE, 1, cell_dims,
      unit_att, coord_unit, file_is_new);
  }
  if (cell_global_ids_available)
    def_var(
      filename, ncid, gid_var_name, NC_INT, 1, cell_dims, NULL, NULL, file_is_new);
  if (core_cell_mask_available)
    def_var(
      filename, ncid, cmk_var_name, NC_INT, 1, cell_dims, NULL, NULL, file_is_new);
  def_var(
    filename, ncid, rnk_var_name, NC_INT, 1, cell_dims, NULL, NULL, file_is_new);

  time_t now = time(NULL);
  char str_now_UTC[32];
  strftime(
    str_now_UTC, sizeof(str_now_UTC), "%Y-%m-%dT%H:%M:%SZ",
    gmtime(&now));
  char const * yac_version = YAC_VERSION;
  char const * created_by = "Created by YAC";
  char const * grid_type = "curvilinear";
  YAC_HANDLE_ERROR(
    nc_put_att_text(ncid, NC_GLOBAL, "YAC", strlen(yac_version), yac_version));
  YAC_HANDLE_ERROR(
    nc_put_att_text(ncid, NC_GLOBAL, "title", strlen(created_by), created_by));
  YAC_HANDLE_ERROR(
    nc_put_att_text(ncid, NC_GLOBAL, "description", strlen(created_by), created_by));
  YAC_HANDLE_ERROR(
    nc_put_att_text(ncid, NC_GLOBAL, "grid", strlen(grid_type), grid_type));
  YAC_HANDLE_ERROR(
    nc_put_att_text(ncid, NC_GLOBAL, "timeStamp", strlen(str_now_UTC), str_now_UTC));

  // end definition
  YAC_HANDLE_ERROR(nc_enddef(ncid));

  // close file
  YAC_HANDLE_ERROR(nc_close(ncid));
}

#endif // YAC_NETCDF_ENABLED

void yac_basic_grid_to_file_parallel(
  struct yac_basic_grid * grid, char const * filename, MPI_Comm comm) {

#ifndef YAC_NETCDF_ENABLED

  UNUSED(grid);
  UNUSED(filename);
  UNUSED(comm);

  die(
    "ERROR(yac_basic_grid_to_file_parallel): "
    "YAC is built without the NetCDF support");
#else

  int comm_rank, comm_size;
  yac_mpi_call(MPI_Comm_rank(comm, &comm_rank), comm);
  yac_mpi_call(MPI_Comm_size(comm, &comm_size), comm);

  { // consistency check
    int filename_len = (int)strlen(filename) + 1;
    yac_mpi_call(MPI_Bcast(&filename_len, 1, MPI_INT, 0, comm), comm);
    char * filename_buffer =
      xmalloc((size_t)filename_len * sizeof(*filename_buffer));
    if (comm_rank == 0) strcpy(filename_buffer, filename);
    yac_mpi_call(
      MPI_Bcast(filename_buffer, filename_len, MPI_CHAR, 0, comm), comm);

    YAC_ASSERT_F(
      !strcmp(filename, filename_buffer),
      "ERROR(yac_basic_grid_to_file_parallel): "
      "inconsistent filename (\"%s\" on rank 0 != \"%s\" on rank %d)",
      filename_buffer, filename, comm_rank);
    free(filename_buffer);
  }

  struct yac_field_data * cell_field =
    yac_basic_grid_get_field_data(grid, YAC_LOC_CELL);
  yac_const_coordinate_pointer cell_center_coords =
    ((cell_field != NULL) &&
     (yac_field_data_get_coordinates_count(cell_field) > 0))?
      yac_field_data_get_coordinates_data(cell_field, 0):NULL;
  struct yac_basic_grid_data * grid_data = yac_basic_grid_get_data(grid);

  uint64_t local_cell_count = (uint64_t)grid->data.num_cells;
  uint64_t * global_cell_counts =
    xmalloc((size_t)comm_size * sizeof(*global_cell_counts));
  int max_num_vertices_per_cell = 0;
  int cell_center_coords_avaiable =
    (grid_data->num_cells == 0) || (cell_center_coords != NULL);
  int cell_global_ids_available =
    (grid_data->num_cells == 0) || (grid_data->cell_ids != NULL);
  int core_cell_mask_available =
    (grid_data->num_cells == 0) || (grid_data->core_cell_mask != NULL);

  for (size_t i = 0; i < grid->data.num_cells; ++i)
    if (grid->data.num_vertices_per_cell[i] > max_num_vertices_per_cell)
      max_num_vertices_per_cell = grid->data.num_vertices_per_cell[i];

  yac_mpi_call(
    MPI_Allgather(
      &local_cell_count, 1, MPI_UINT64_T,
      global_cell_counts, 1, MPI_UINT64_T, comm), comm);
  yac_mpi_call(
    MPI_Allreduce(
      MPI_IN_PLACE, &max_num_vertices_per_cell, 1, MPI_INT, MPI_MAX, comm),
    comm);
  int ints[3] = {cell_center_coords_avaiable,
                 cell_global_ids_available,
                 core_cell_mask_available};
  yac_mpi_call(
    MPI_Allreduce(MPI_IN_PLACE, ints, 3, MPI_INT, MPI_MIN, comm), comm);
  cell_center_coords_avaiable = ints[0];
  cell_global_ids_available = ints[1];
  core_cell_mask_available = ints[2];

  size_t global_cell_count = 0;
  for (int i = 0; i < comm_size; ++i)
    global_cell_count += (size_t)(global_cell_counts[i]);

  YAC_ASSERT_F(
    global_cell_count > 0,
    "ERROR(yac_basic_grid_to_file_parallel): grid \"%s\" has no cells",
    grid->name);

  // create the grid file
  if (comm_rank == 0)
    create_grid_file(
      filename, grid->name, global_cell_count, max_num_vertices_per_cell,
      cell_center_coords_avaiable, cell_global_ids_available,
      core_cell_mask_available);
  yac_mpi_call(MPI_Barrier(comm), comm);

  // determine processes that will do output
  int io_flag;
  int * io_ranks;
  int num_io_ranks;
  yac_get_io_ranks(comm, &io_flag, &io_ranks, &num_io_ranks);

  // redistribute data
  int * int_buffer = xcalloc(4 * (size_t)comm_size, sizeof(*int_buffer));
  int * sendcounts = int_buffer + 0 * comm_size;
  int * recvcounts = int_buffer + 1 * comm_size;
  int * sdispls =    int_buffer + 2 * comm_size;
  int * rdispls =    int_buffer + 3 * comm_size;

  size_t io_start_idx, io_end_idx = 0;
  size_t global_count_accu = 0;
  size_t recv_count = 0;
  size_t curr_send_size = 0;
  for (int io_rank_idx = 0, curr_rank = 0; io_rank_idx < num_io_ranks;
       ++io_rank_idx) {

    int curr_io_rank = io_ranks[io_rank_idx];
    io_start_idx = io_end_idx;
    io_end_idx =
      (size_t)(
        ((long long)(io_rank_idx + 1) * (long long)global_cell_count)/
        (long long)num_io_ranks);
    size_t io_size = io_end_idx - io_start_idx;
    size_t curr_recv_size = 0;

    if (curr_io_rank == comm_rank) recv_count = io_size;

    while(global_count_accu < io_end_idx) {

      size_t sendcount =
        MIN(
          (global_count_accu + global_cell_counts[curr_rank]) - (io_start_idx + curr_recv_size),
          io_size - curr_recv_size);

      if (curr_rank == comm_rank) {
        sendcounts[curr_io_rank] = (int)sendcount;
        sdispls[curr_io_rank] = (int)curr_send_size;
        curr_send_size += sendcount;
      }

      if (curr_io_rank == comm_rank) {
        recvcounts[curr_rank] = (int)sendcount;
        rdispls[curr_rank] = (int)curr_recv_size;
      }

      if ((global_count_accu + global_cell_counts[curr_rank]) <=
          io_end_idx) {
        global_count_accu += global_cell_counts[curr_rank];
        curr_rank++;
        curr_send_size = 0;
      }

      curr_recv_size += sendcount;

      if (curr_recv_size >= io_size) break;
    }
  }

  free(global_cell_counts);

  // set up send and recv buffers
  double * sendbuffer_dble =
    xmalloc(
      (2 * grid->data.num_cells * (size_t)max_num_vertices_per_cell + // corner lon/lat
      2 * (cell_center_coords_avaiable?grid->data.num_cells:0)) * // cell lon/lat
      sizeof(*sendbuffer_dble));
  double * recvbuffer_dble =
    xmalloc(
      (2 * recv_count * (size_t)max_num_vertices_per_cell + // corner lon/lat
      2 * (cell_center_coords_avaiable?recv_count:0)) * // cell lon/lat
      sizeof(*recvbuffer_dble));
  int * sendbuffer_int =
    xmalloc(
      (cell_global_ids_available + core_cell_mask_available) *
      grid->data.num_cells * sizeof(*sendbuffer_int));
  int * recvbuffer_int =
    xmalloc(
      (cell_global_ids_available + core_cell_mask_available) *
      recv_count * sizeof(*recvbuffer_int));

  // fill send buffer
  for (size_t i = 0, offset = 0; i < grid->data.num_cells; ++i) {
    int j;
    for (j = 0; j < grid->data.num_vertices_per_cell[i]; ++j) {
      XYZtoLL(
        grid->data.vertex_coordinates[
          grid->data.cell_to_vertex[
            grid->data.cell_to_vertex_offsets[i] + j]],
        sendbuffer_dble + offset + 2 * j + 0,
        sendbuffer_dble + offset + 2 * j + 1);
      sendbuffer_dble[offset + 2 * j + 0] /= YAC_RAD;
      sendbuffer_dble[offset + 2 * j + 1] /= YAC_RAD;
    }
    for (; j < max_num_vertices_per_cell; ++j) {
      sendbuffer_dble[offset + 2 * j + 0] = sendbuffer_dble[offset + 0];
      sendbuffer_dble[offset + 2 * j + 1] = sendbuffer_dble[offset + 1];
    }
    offset += 2 * (size_t)max_num_vertices_per_cell;

    if (cell_center_coords_avaiable) {
      XYZtoLL(
        cell_center_coords[i],
        sendbuffer_dble + offset + 0,
        sendbuffer_dble + offset + 1);
      sendbuffer_dble[offset + 0] /= YAC_RAD;
      sendbuffer_dble[offset + 1] /= YAC_RAD;
      offset += 2;
    }
  }

  int ints_mult = cell_global_ids_available + core_cell_mask_available;

  if (ints_mult > 0) {

    for (size_t i = 0; i < grid->data.num_cells; ++i) {
      if (cell_global_ids_available)
        sendbuffer_int[i * ints_mult + 0] = (int)(grid_data->cell_ids[i]);
      if (core_cell_mask_available)
        sendbuffer_int[i * ints_mult + cell_global_ids_available] =
          (int)(grid_data->core_cell_mask[i]);
    }

    for (int i = 0; i < comm_size; ++i) {
      sendcounts[i] *= ints_mult;
      recvcounts[i] *= ints_mult;
      sdispls[i] *= ints_mult;
      rdispls[i] *= ints_mult;
    }

    // redistribute grid data
    yac_mpi_call(
      MPI_Alltoallv(
        sendbuffer_int, sendcounts, sdispls, MPI_INT,
        recvbuffer_int, recvcounts, rdispls, MPI_INT, comm),
      comm);

    for (int i = 0; i < comm_size; ++i) {
      sendcounts[i] /= ints_mult;
      recvcounts[i] /= ints_mult;
      sdispls[i] /= ints_mult;
      rdispls[i] /= ints_mult;
    }
  }

  for (int i = 0,
       mult = 2 * (max_num_vertices_per_cell + cell_center_coords_avaiable);
       i < comm_size; ++i) {
    sendcounts[i] *= mult;
    recvcounts[i] *= mult;
    sdispls[i] *= mult;
    rdispls[i] *= mult;
  }

  // redistribute grid data
  yac_mpi_call(
    MPI_Alltoallv(
      sendbuffer_dble, sendcounts, sdispls, MPI_DOUBLE,
      recvbuffer_dble, recvcounts, rdispls, MPI_DOUBLE, comm), comm);

  for (int i = 0,
       mult = 2 * (max_num_vertices_per_cell + cell_center_coords_avaiable);
       i < comm_size; ++i) {
    sendcounts[i] /= mult;
    recvcounts[i] /= mult;
    sdispls[i] /= mult;
    rdispls[i] /= mult;
  }

  free(sendbuffer_dble);
  free(sendbuffer_int);

  if (io_flag) {

    int io_rank_idx;
    for (io_rank_idx = 0; io_rank_idx < num_io_ranks; ++io_rank_idx)
      if (io_ranks[io_rank_idx] == comm_rank) break;

    size_t io_start_idx =
      (size_t)(
        ((long long)(io_rank_idx) * (long long)global_cell_count)/
        (long long)num_io_ranks);

    int ncid;

    // open weight file
    yac_nc_open(filename, NC_WRITE | NC_SHARE, &ncid);

    size_t grid_name_len = strlen(grid->name) + 1;

    char cla_var_name[4 + grid_name_len];
    char clo_var_name[4 + grid_name_len];
    char lat_var_name[4 + grid_name_len];
    char lon_var_name[4 + grid_name_len];
    char gid_var_name[4 + grid_name_len];
    char cmk_var_name[4 + grid_name_len];
    char rnk_var_name[4 + grid_name_len];

    snprintf(cla_var_name, 4 + grid_name_len, "%s.cla", grid->name);
    snprintf(clo_var_name, 4 + grid_name_len, "%s.clo", grid->name);
    snprintf(lat_var_name, 4 + grid_name_len, "%s.lat", grid->name);
    snprintf(lon_var_name, 4 + grid_name_len, "%s.lon", grid->name);
    snprintf(gid_var_name, 4 + grid_name_len, "%s.gid", grid->name);
    snprintf(cmk_var_name, 4 + grid_name_len, "%s.cmk", grid->name);
    snprintf(rnk_var_name, 4 + grid_name_len, "%s.rnk", grid->name);

    int cla_var_id, clo_var_id;
    yac_nc_inq_varid(ncid, cla_var_name, &cla_var_id);
    yac_nc_inq_varid(ncid, clo_var_name, &clo_var_id);

    double * write_buffer_dble =
      xmalloc(recv_count * (size_t)max_num_vertices_per_cell * sizeof(*write_buffer_dble));

    size_t start[2] = {io_start_idx, 0};
    size_t count[2] = {recv_count, max_num_vertices_per_cell};

    // clo data
    for (size_t i = 0, offset = 0; i < recv_count; ++i) {
      for (size_t j = 0; j < (size_t)max_num_vertices_per_cell; ++j)
        write_buffer_dble[i * (size_t)max_num_vertices_per_cell + j] =
          recvbuffer_dble[offset + 2 * j + 0];
      offset += 2 * (max_num_vertices_per_cell + cell_center_coords_avaiable);
    }
    YAC_HANDLE_ERROR(
      nc_put_vara_double(ncid, clo_var_id, start, count, write_buffer_dble));

    // cla data
    for (size_t i = 0, offset = 0; i < recv_count; ++i) {
      for (size_t j = 0; j < (size_t)max_num_vertices_per_cell; ++j)
        write_buffer_dble[i * (size_t)max_num_vertices_per_cell + j] =
          recvbuffer_dble[offset + 2 * j + 1];
      offset += 2 * (max_num_vertices_per_cell + cell_center_coords_avaiable);
    }
    YAC_HANDLE_ERROR(
      nc_put_vara_double(ncid, cla_var_id, start, count, write_buffer_dble));

    if (cell_center_coords_avaiable) {

      int lat_var_id, lon_var_id;
      yac_nc_inq_varid(ncid, lat_var_name, &lat_var_id);
      yac_nc_inq_varid(ncid, lon_var_name, &lon_var_id);

      // lon data
      for (size_t i = 0, offset = 2 * max_num_vertices_per_cell;
           i < recv_count; ++i) {
        write_buffer_dble[i] = recvbuffer_dble[offset];
        offset += 2 * (max_num_vertices_per_cell + 1);
      }
      YAC_HANDLE_ERROR(
        nc_put_vara_double(ncid, lon_var_id, start, count, write_buffer_dble));

      // lat data
      for (size_t i = 0, offset = 2 * max_num_vertices_per_cell + 1;
           i < recv_count; ++i) {
        write_buffer_dble[i] = recvbuffer_dble[offset];
        offset += 2 * (max_num_vertices_per_cell + 1);
      }
      YAC_HANDLE_ERROR(
        nc_put_vara_double(ncid, lat_var_id, start, count, write_buffer_dble));
    }

    free(write_buffer_dble);

    int * write_buffer_int = xmalloc(recv_count * sizeof(*write_buffer_int));

    if (cell_global_ids_available) {

      int gid_var_id;
      yac_nc_inq_varid(ncid, gid_var_name, &gid_var_id);

      // global cell id data
      for (size_t i = 0; i < recv_count; ++i)
        write_buffer_int[i] = recvbuffer_int[i * ints_mult + 0];
      YAC_HANDLE_ERROR(
        nc_put_vara_int(ncid, gid_var_id, start, count, write_buffer_int));
    }

    if (core_cell_mask_available) {

      int cmk_var_id;
      yac_nc_inq_varid(ncid, cmk_var_name, &cmk_var_id);

      // global core cell mask data
      for (size_t i = 0; i < recv_count; ++i)
        write_buffer_int[i] =
          recvbuffer_int[i * ints_mult + cell_global_ids_available];
      YAC_HANDLE_ERROR(
        nc_put_vara_int(ncid, cmk_var_id, start, count, write_buffer_int));
    }

    {
      int rnk_var_id;
      yac_nc_inq_varid(ncid, rnk_var_name, &rnk_var_id);

      // owner ranks
      size_t i = 0;
      for (int rank = 0; rank < comm_size; ++rank)
        for (int j = 0; j < recvcounts[rank]; ++j, ++i)
          write_buffer_int[i] = rank;
      YAC_HANDLE_ERROR(
        nc_put_vara_int(ncid, rnk_var_id, start, count, write_buffer_int));
    }

    free(write_buffer_int);

    // close weight file
    YAC_HANDLE_ERROR(nc_close(ncid));
  }

  free(int_buffer);
  free(recvbuffer_int);
  free(recvbuffer_dble);

  free(io_ranks);

  // ensure that the writing of the weight file is complete
  yac_mpi_call(MPI_Barrier(comm), comm);

#endif // YAC_NETCDF_ENABLED
}
