// Copyright (c) 2024 The YAC Authors
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdlib.h>

#include "tests.h"
#include "test_common.h"
#include "basic_grid.h"
#include "io_utils.h"
#include "geometry.h"
#include "grid_file_common.h"

#include <mpi.h>
#include <netcdf.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

static void delete_file(char const * filename);

int main(void) {

  MPI_Init(NULL, NULL);

  int comm_rank, comm_size;
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Barrier(MPI_COMM_WORLD);

  enum {NUM_PROCS = 3};
  if (comm_size != NUM_PROCS) {
    PUT_ERR("ERROR: wrong number of processes");
    xt_finalize();
    MPI_Finalize();
    return TEST_EXIT_CODE;
  }

  {
    struct {
      size_t nbr_vertices[NUM_PROCS][2];
      double lon_vertices[NUM_PROCS][6];
      double lat_vertices[NUM_PROCS][6];
      yac_int global_ids[NUM_PROCS][25];
      int core_cell_mask[NUM_PROCS][25];
    } decomp[] =
      {{.nbr_vertices = {{2,5},{5,5},{6,2}},
        .lon_vertices = {{0,1},{1,2,3,4,5},{0,1,2,3,4,5}},
        .lat_vertices = {{0,1,2,3,4},{0,1,2,3,4},{4,5}},
        .global_ids = {{0,5,10,15},
                       {1,2,3,4, 6,7,8,9, 11,12,13,14, 16,17,18,19},
                       {20,21,22,23,24}},
        .core_cell_mask = {{1,0,1,0},
                           {0,1,0,1, 1,0,1,0, 0,1,0,1, 1,0,1,0},
                           {1,0,1,0,1}}},
       {.nbr_vertices = {{6,6},{0,0},{0,0}},
        .lon_vertices = {{0,1,2,3,4,5},{-1},{-1}},
        .lat_vertices = {{0,1,2,3,4,5},{-1},{-1}},
        .global_ids = {{0,1,2,3,4,
                        5,6,7,8,9,
                        10,11,12,13,14,
                        15,16,17,18,19,
                        20,21,22,23,24}, {-1}, {-1}},
        .core_cell_mask = {{1,0,1,0,1,
                            0,1,0,1,0,
                            1,0,1,0,1,
                            0,1,0,1,0,
                            1,0,1,0,1}, {-1}, {-1}}},
       {.nbr_vertices = {{3,3},{6,6},{0,0}},
        .lon_vertices = {{0,1,2},{0,1,2,3,4,5},{-1}},
        .lat_vertices = {{0,1,2},{0,1,2,3,4,5},{-1}},
        .global_ids = {{0,1,
                        5,6},
                       {0,1,2,3,4,
                        5,6,7,8,9,
                        10,11,12,13,14,
                        15,16,17,18,19,
                        20,21,22,23,24}, {-1}},
        .core_cell_mask = {{1,0,
                            0,1},
                           {1,0,1,0,1,
                            0,1,0,1,0,
                            1,0,1,0,1,
                            0,1,0,1,0,
                            1,0,1,0,1}, {-1}}}};
    enum {NUM_DECOMP = sizeof(decomp) / sizeof(decomp[0])};
    char const * grid_filename = "test_basic_grid_parallel.nc";

    if (yac_file_exists(grid_filename)) delete_file(grid_filename);

    // for all grid decompositions
    for (size_t decomp_idx = 0; decomp_idx < NUM_DECOMP; ++decomp_idx) {

      for (int with_core_cell_mask = 0; with_core_cell_mask <= 1; ++with_core_cell_mask) {

        for (int with_cell_centers = 0; with_cell_centers <= 1; ++with_cell_centers) {

          for (int with_global_ids = 0; with_global_ids <= 1; ++with_global_ids) {


            // set up distributed grid
            int cyclic[2] = {0,0};

            struct yac_basic_grid * grid;
            char const * grid_name = "test_grid";
            if (decomp[decomp_idx].nbr_vertices[comm_rank][0] *
                decomp[decomp_idx].nbr_vertices[comm_rank][1] > 0) {
              struct yac_basic_grid_data grid_data =
                yac_generate_basic_grid_data_reg_2d_deg(
                  decomp[decomp_idx].nbr_vertices[comm_rank], cyclic,
                  decomp[decomp_idx].lon_vertices[comm_rank],
                  decomp[decomp_idx].lat_vertices[comm_rank]);
              grid = yac_basic_grid_new(grid_name, grid_data);
            } else {
              grid = yac_basic_grid_empty_new(grid_name);
            }

            // compute cell centers
            struct yac_basic_grid_data * basic_grid_data =
              yac_basic_grid_get_data(grid);
            yac_coordinate_pointer cell_center_coords =
              malloc(basic_grid_data->num_cells * sizeof(*cell_center_coords));
            for (size_t i = 0; i < basic_grid_data->num_cells; ++i) {
              cell_center_coords[i][0] = 0.0;
              cell_center_coords[i][1] = 0.0;
              cell_center_coords[i][2] = 0.0;
              for (int j = 0; j < basic_grid_data->num_vertices_per_cell[i]; ++j)
                for (size_t k = 0; k < 3; ++k)
                  cell_center_coords[i][k] +=
                    basic_grid_data->vertex_coordinates[
                      basic_grid_data->cell_to_vertex[
                        basic_grid_data->cell_to_vertex_offsets[i] + j]][k];
              normalise_vector(cell_center_coords[i]);
            }
            if (with_cell_centers && (basic_grid_data->num_cells > 0))
              yac_basic_grid_add_coordinates(
                grid, YAC_LOC_CELL, cell_center_coords, basic_grid_data->num_cells);
            free(cell_center_coords);
            if (with_global_ids && (basic_grid_data->num_cells > 0))
              basic_grid_data->cell_ids =
                to_pointer(
                  decomp[decomp_idx].global_ids[comm_rank],
                  basic_grid_data->num_cells * sizeof(decomp[0].global_ids[0][0]));
            if (with_core_cell_mask && (basic_grid_data->num_cells > 0))
              basic_grid_data->core_cell_mask =
                to_pointer(
                  decomp[decomp_idx].core_cell_mask[comm_rank],
                  basic_grid_data->num_cells * sizeof(decomp[0].core_cell_mask[0][0]));

            // set up io configuration
            clear_yac_io_env();
            setenv("YAC_IO_RANK_LIST", "0,2", 1);
            setenv("YAC_IO_MAX_NUM_RANKS_PER_NODE", "3", 1);

            // write distributed grid to file
            yac_basic_grid_to_file_parallel(grid, grid_filename, MPI_COMM_WORLD);

            yac_basic_grid_delete(grid);

            // check generated grid file

            if (comm_rank == 0) {
              enum {NUM_CELLS = 25, NUM_CRN = 4};
              double ref_cla[NUM_CELLS][NUM_CRN] =
                {{0,0,1,1},{0,0,1,1},{0,0,1,1},{0,0,1,1},{0,0,1,1},
                {1,1,2,2},{1,1,2,2},{1,1,2,2},{1,1,2,2},{1,1,2,2},
                {2,2,3,3},{2,2,3,3},{2,2,3,3},{2,2,3,3},{2,2,3,3},
                {3,3,4,4},{3,3,4,4},{3,3,4,4},{3,3,4,4},{3,3,4,4},
                {4,4,5,5},{4,4,5,5},{4,4,5,5},{4,4,5,5},{4,4,5,5}};
              double ref_clo[NUM_CELLS][NUM_CRN] =
                {{0,1,1,0},{1,2,2,1},{2,3,3,2},{3,4,4,3},{4,5,5,4},
                {0,1,1,0},{1,2,2,1},{2,3,3,2},{3,4,4,3},{4,5,5,4},
                {0,1,1,0},{1,2,2,1},{2,3,3,2},{3,4,4,3},{4,5,5,4},
                {0,1,1,0},{1,2,2,1},{2,3,3,2},{3,4,4,3},{4,5,5,4},
                {0,1,1,0},{1,2,2,1},{2,3,3,2},{3,4,4,3},{4,5,5,4}};
              double ref_lat[NUM_CELLS] =
                {0.5,0.5,0.5,0.5,0.5,
                1.5,1.5,1.5,1.5,1.5,
                2.5,2.5,2.5,2.5,2.5,
                3.5,3.5,3.5,3.5,3.5,
                4.5,4.5,4.5,4.5,4.5};
              double ref_lon[NUM_CELLS] =
                {0.5,1.5,2.5,3.5,4.5,
                0.5,1.5,2.5,3.5,4.5,
                0.5,1.5,2.5,3.5,4.5,
                0.5,1.5,2.5,3.5,4.5,
                0.5,1.5,2.5,3.5,4.5};
              int ref_global_ids[NUM_CELLS] =
                {0,1,2,3,4,
                5,6,7,8,9,
                10,11,12,13,14,
                15,16,17,18,19,
                20,21,22,23,24};
              int ref_core_cell_mask[NUM_CELLS] =
                {1,0,1,0,1,
                0,1,0,1,0,
                1,0,1,0,1,
                0,1,0,1,0,
                1,0,1,0,1};

              check_grid_file(
                grid_filename, grid_name, NUM_CELLS, NUM_CRN,
                &ref_cla[0][0], &ref_clo[0][0],
                with_cell_centers?ref_lat:NULL,
                with_cell_centers?ref_lon:NULL,
                with_global_ids?ref_global_ids:NULL,
                with_core_cell_mask?ref_core_cell_mask:NULL);
            } // comm_rank == 0
          } // with global ids
          delete_file(grid_filename);
        } // with cell centers
      } // with core mask
      delete_file(grid_filename);
    } // decomp
  }

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static void delete_file(char const * filename) {

  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Barrier(MPI_COMM_WORLD);
  if (rank == 0) unlink(filename);
  MPI_Barrier(MPI_COMM_WORLD);
}
